from ._factories import (
    CacheModifiedInputIterator,
    ConstantIterator,
    CountingIterator,
    DiscardIterator,
    PermutationIterator,
    ReverseIterator,
    TransformIterator,
    TransformOutputIterator,
    ZipIterator,
)

__all__ = [
    "CacheModifiedInputIterator",
    "ConstantIterator",
    "CountingIterator",
    "DiscardIterator",
    "PermutationIterator",
    "ReverseIterator",
    "TransformIterator",
    "TransformOutputIterator",
    "ZipIterator",
]



# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

import ctypes
from typing import Callable

import numpy as np
from numba import cuda

from .. import types
from .._bindings import IteratorState
from .._caching import CachableFunction, cache_with_registered_key_functions
from .._intrinsics import load_cs
from .._utils.protocols import (
    compute_c_contiguous_strides_in_bytes,
    get_data_pointer,
    get_dtype,
    get_shape,
)
from ..typing import DeviceArrayLike


class IteratorKind:
    # The `.kind` of an iterator encapsulates additional metadata about the iterator,
    # analogous to the `.dtype` of a NumPy array.
    def __init__(
        self, value_type: types.TypeDescriptor, state_type: types.TypeDescriptor
    ):
        self.value_type = value_type
        self.state_type = state_type

    def __repr__(self):
        return (
            f"{self.__class__.__name__}[{str(self.value_type), str(self.state_type)}]"
        )

    def __eq__(self, other):
        return (
            type(self) is type(other)
            and self.value_type == other.value_type
            and self.state_type == other.state_type
        )

    def __hash__(self):
        return hash((type(self), self.value_type, self.state_type))


class IteratorBase:
    """
    An Iterator is a wrapper around a state pointer, and must define the following:

    - an `advance` property returning a (static) method which receives the
      state pointer and performs an action that advances the state pointer by the offset
      `distance` (returns nothing).
    - `input_dereference` and `output_dereference` properties that return
      (static) methods for reading from and writing to the state pointer respectively.

    Iterators are not meant to be used directly. They are constructed and passed
    to algorithms which internally invoke their methods.

    The `advance`, `input_dereference`, and `output_dereference` must be compilable
    to device code by numba.
    """

    def __init__(
        self,
        kind: IteratorKind,
        cvalue,
        state_type: types.TypeDescriptor,
        value_type: types.TypeDescriptor,
    ):
        """
        Parameters
        ----------
        cvalue
          A ctypes type representing the object pointed to by the iterator.
        state_type
          A TypeDescriptor representing the state type
        value_type
          A TypeDescriptor representing the value type
        """
        self._kind = kind
        self.cvalue = cvalue
        self.state_type = state_type
        self.value_type = value_type
        self._state = IteratorState(self.cvalue)

    @property
    def kind(self):
        return self._kind

    @property
    def state(self) -> IteratorState:
        return self._state

    @property
    def advance(state):
        raise NotImplementedError("Subclasses must override advance property")

    @property
    def input_dereference(state):
        return None

    @property
    def output_dereference(state):
        return None

    @property
    def host_advance(self):
        return None

    @property
    def children(self):
        return ()

    @property
    def is_input_iterator(self) -> bool:
        return self.input_dereference is not None

    @property
    def is_output_iterator(self) -> bool:
        return self.output_dereference is not None

    def __add__(self, offset: int):
        # add the offset to the iterator's state, and return a new iterator
        # with the new state.
        res = type(self).__new__(type(self))
        res.state_type = self.state_type
        res.value_type = self.value_type
        res._kind = self._kind
        res.cvalue = type(self.cvalue)(self.cvalue.value + offset)
        res._state = IteratorState(res.cvalue)

        return res


class RawPointerKind(IteratorKind):
    pass


class RawPointer(IteratorBase):
    def __init__(self, ptr: int, value_type: types.TypeDescriptor, obj: object):
        """
        Args:
            ptr: Device pointer address
            value_type: A TypeDescriptor for the element type
            obj: Reference to the underlying object (to prevent garbage collection)
        """

        cvalue = ctypes.c_void_p(ptr)
        # state_type is a pointer to value_type
        state_type = value_type.pointer()
        self.obj = obj
        kind = RawPointerKind(value_type, state_type)

        super().__init__(
            kind=kind,
            cvalue=cvalue,
            state_type=state_type,
            value_type=value_type,
        )

    @property
    def host_advance(self):
        return self._advance

    @property
    def advance(self):
        return self._advance

    @staticmethod
    def _advance(state, distance):
        state[0] = state[0] + distance

    @staticmethod
    def input_dereference(state, result):
        result[0] = state[0][0]

    @staticmethod
    def output_dereference(state, x):
        state[0][0] = x


def pointer(container) -> RawPointer:
    """Create a RawPointer iterator from a device array."""

    return RawPointer(
        container.__cuda_array_interface__["data"][0],
        types.from_numpy_dtype(get_dtype(container)),
        container,
    )


class CacheModifiedPointerKind(IteratorKind):
    pass


class CacheModifiedPointer(IteratorBase):
    def __init__(self, ptr: int, value_type: types.TypeDescriptor):
        """
        Args:
            ptr: Device pointer address
            value_type: A TypeDescriptor for the element type
        """
        cvalue = ctypes.c_void_p(ptr)
        state_type = value_type.pointer()
        kind = CacheModifiedPointerKind(value_type, state_type)
        super().__init__(
            kind=kind,
            cvalue=cvalue,
            state_type=state_type,
            value_type=value_type,
        )

    @property
    def host_advance(self):
        return self._advance

    @property
    def advance(self):
        return self._advance

    @property
    def input_dereference(self):
        return self._input_dereference

    @property
    def output_dereference(self):
        raise AttributeError(
            "CacheModifiedPointer cannot be used as an output iterator"
        )

    @staticmethod
    def _advance(state, distance):
        state[0] = state[0] + distance

    @staticmethod
    def _input_dereference(state, result):
        result[0] = load_cs(state[0])


class ConstantIteratorKind(IteratorKind):
    pass


class ConstantIterator(IteratorBase):
    def __init__(self, value: np.number):
        value_type = types.from_numpy_dtype(value.dtype)
        cvalue = types.to_ctypes_type(value_type)(value)
        # state_type is the value itself (not a pointer)
        state_type = value_type
        kind = ConstantIteratorKind(value_type, state_type)
        super().__init__(
            kind=kind,
            cvalue=cvalue,
            state_type=state_type,
            value_type=value_type,
        )

    @property
    def host_advance(self):
        return self._advance

    @property
    def advance(self):
        return self._advance

    @property
    def input_dereference(self):
        return self._input_dereference

    @property
    def output_dereference(self):
        raise AttributeError("ConstantIterator cannot be used as an output iterator")

    @staticmethod
    def _advance(state, distance):
        pass

    @staticmethod
    def _input_dereference(state, result):
        result[0] = state[0]


class CountingIteratorKind(IteratorKind):
    pass


class CountingIterator(IteratorBase):
    def __init__(self, value: np.number):
        value_type = types.from_numpy_dtype(value.dtype)
        cvalue = types.to_ctypes_type(value_type)(value)
        # state_type is the value itself (not a pointer)
        state_type = value_type
        kind = CountingIteratorKind(value_type, state_type)
        super().__init__(
            kind=kind,
            cvalue=cvalue,
            state_type=state_type,
            value_type=value_type,
        )

    @property
    def host_advance(self):
        return self._advance

    @property
    def advance(self):
        return self._advance

    @property
    def input_dereference(self):
        return self._input_dereference

    @property
    def output_dereference(self):
        raise AttributeError("CountingIterator cannot be used as an output iterator")

    @staticmethod
    def _advance(state, distance):
        state[0] = state[0] + distance

    @staticmethod
    def _input_dereference(state, result):
        result[0] = state[0]


class DiscardIteratorKind(IteratorKind):
    pass


class DiscardIterator(IteratorBase):
    def __init__(self, reference_iterator=None):
        from .._utils.temp_storage_buffer import TempStorageBuffer

        if reference_iterator is None:
            reference_iterator = TempStorageBuffer(1)

        if hasattr(reference_iterator, "__cuda_array_interface__"):
            iter = RawPointer(
                reference_iterator.__cuda_array_interface__["data"][0],
                types.from_numpy_dtype(get_dtype(reference_iterator)),
                reference_iterator,
            )
        else:
            iter = reference_iterator

        super().__init__(
            kind=DiscardIteratorKind(iter.value_type, iter.state_type),
            cvalue=iter.cvalue,
            state_type=iter.state_type,
            value_type=iter.value_type,
        )

    @property
    def host_advance(self):
        return self._advance

    @property
    def advance(self):
        return self._advance

    @property
    def input_dereference(self):
        return self._dereference

    @property
    def output_dereference(self):
        return self._dereference

    @staticmethod
    def _advance(state, distance):
        pass

    @staticmethod
    def _dereference(state, result):
        pass


class ReverseIteratorKind(IteratorKind):
    pass


def make_reverse_iterator(it: DeviceArrayLike | IteratorBase):
    if not hasattr(it, "__cuda_array_interface__") and not isinstance(it, IteratorBase):
        raise NotImplementedError(
            f"Reverse iterator is not implemented for type {type(it)}"
        )

    if hasattr(it, "__cuda_array_interface__"):
        last_element_ptr = _get_last_element_ptr(it)
        it = RawPointer(
            last_element_ptr,
            types.from_numpy_dtype(get_dtype(it)),
            it,
        )

    it_advance = cuda.jit(it.advance, device=True)
    it_input_dereference = (
        cuda.jit(it.input_dereference, device=True)
        if hasattr(it, "input_dereference")
        else None
    )
    it_output_dereference = (
        cuda.jit(it.output_dereference, device=True)
        if hasattr(it, "output_dereference")
        else None
    )

    class ReverseIterator(IteratorBase):
        def __init__(self, it):
            self._it = it
            kind = ReverseIteratorKind(it.value_type, it.state_type)
            super().__init__(
                kind=kind,
                cvalue=it.cvalue,
                state_type=it.state_type,
                value_type=it.value_type,
            )

        @property
        def host_advance(self):
            return self._advance

        @property
        def children(self):
            return (self._it,)

        @property
        def advance(self):
            return self._advance

        @property
        def input_dereference(self):
            if it_input_dereference is None:
                raise AttributeError("This iterator is not an input iterator")
            return ReverseIterator._input_dereference

        @property
        def output_dereference(self):
            if it_output_dereference is None:
                raise AttributeError("This iterator is not an output iterator")
            return ReverseIterator._output_dereference

        @staticmethod
        def _advance(state, distance):
            return it_advance(state, -distance)

        @staticmethod
        def _input_dereference(state, result):
            it_input_dereference(state, result)

        @staticmethod
        def _output_dereference(state, x):
            it_output_dereference(state, x)

    return ReverseIterator(it)


class TransformIteratorKind(IteratorKind):
    def __init__(
        self, underlying_it_kind: IteratorKind, op: CachableFunction, io_kind: str
    ):
        self.underlying_it_kind = underlying_it_kind
        self.op = op
        self.io_kind = io_kind

    def __repr__(self):
        return f"TransformIteratorKind({self.underlying_it_kind}, {self.op})"

    def __eq__(self, other):
        return (
            type(self) is type(other)
            and self.underlying_it_kind == other.underlying_it_kind
            and self.op == other.op
        )

    def __hash__(self):
        return hash((type(self), self.underlying_it_kind, self.op))


def make_transform_iterator(it: IteratorBase, op: Callable, io_kind: str):
    if hasattr(it, "__cuda_array_interface__"):
        it = pointer(it)

    it_host_advance = it.host_advance
    it_advance = cuda.jit(it.advance, device=True)

    it_input_dereference = (
        cuda.jit(it.input_dereference, device=True)
        if hasattr(it, "input_dereference")
        else None
    )
    it_output_dereference = (
        cuda.jit(it.output_dereference, device=True)
        if hasattr(it, "output_dereference")
        else None
    )

    alloca_temp_for_underlying_type = None
    if io_kind == "input":
        from .._intrinsics import make_alloca_intrinsic

        # Create alloca intrinsic (accepts TypeDescriptor directly)
        alloca_temp_for_underlying_type = make_alloca_intrinsic(it.value_type)

    op = cuda.jit(op, device=True)

    class TransformIterator(IteratorBase):
        def __init__(self, it: IteratorBase, op, io_kind: str):
            self._it = it
            self._op = CachableFunction(op.py_func)
            state_type = it.state_type

            kind = TransformIteratorKind(it.kind, self._op, io_kind)
            super().__init__(
                kind=kind,
                cvalue=it.cvalue,
                state_type=state_type,
                value_type=it.value_type,
            )

        @property
        def host_advance(self):
            return it_host_advance

        @property
        def children(self):
            return (self._it,)

        @property
        def advance(self):
            return self._advance

        @property
        def input_dereference(self):
            if it_input_dereference is None:
                raise AttributeError("This iterator is not an input iterator")
            return TransformIterator._input_dereference

        @property
        def output_dereference(self):
            if it_output_dereference is None:
                raise AttributeError("This iterator is not an output iterator")
            return TransformIterator._output_dereference

        @staticmethod
        def _advance(state, distance):
            return it_advance(state, distance)

        @staticmethod
        def _input_dereference(state, result):
            # Allocate temporary storage for the deref input type
            temp_ptr = alloca_temp_for_underlying_type()
            # Call underlying iterator's dereference with temp storage
            it_input_dereference(state, temp_ptr)
            # Apply transformation and store in result
            result[0] = op(temp_ptr[0])

        @staticmethod
        def _output_dereference(state, x):
            it_output_dereference(state, op(x))

        def __add__(self, offset: int):
            res = type(self).__new__(type(self))
            res._it = self._it + offset
            res._op = self._op
            res.state_type = self.state_type
            res.value_type = self.value_type
            res._kind = self._kind
            res.cvalue = type(self.cvalue)(self.cvalue.value + offset)
            res._state = IteratorState(res.cvalue)
            return res

    return TransformIterator(it, op, io_kind)


def _get_last_element_ptr(device_array) -> int:
    shape = get_shape(device_array)
    dtype = get_dtype(device_array)

    strides_in_bytes = device_array.__cuda_array_interface__["strides"]
    if strides_in_bytes is None:
        strides_in_bytes = compute_c_contiguous_strides_in_bytes(shape, dtype.itemsize)

    offset_to_last_element = sum(
        (dim_size - 1) * stride for dim_size, stride in zip(shape, strides_in_bytes)
    )

    ptr = get_data_pointer(device_array)
    return ptr + offset_to_last_element


cache_with_registered_key_functions.register(IteratorBase, lambda it: it.kind)



# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

import ctypes

from numba import cuda

from .. import types
from .._caching import cache_with_registered_key_functions
from ._iterators import (
    IteratorBase,
    IteratorKind,
    pointer,
)


class PermutationIteratorKind(IteratorKind):
    def __init__(
        self,
        value_type: types.TypeDescriptor,
        values_kind: IteratorKind,
        indices_kind: IteratorKind,
    ):
        self.values_kind = values_kind
        self.indices_kind = indices_kind

    def __repr__(self):
        return f"PermutationIteratorKind({self.value_type}, {self.values_kind}, {self.indices_kind})"

    def __eq__(self, other):
        return (
            type(self) is type(other)
            and self.values_kind == other.values_kind
            and self.indices_kind == other.indices_kind
        )

    def __hash__(self):
        return hash((type(self), self.values_kind, self.indices_kind))


@cache_with_registered_key_functions
def _generate_advance_and_dereference_methods(values, indices):
    from .._intrinsics import make_alloca_intrinsic, make_struct_field_ptr_intrinsic

    # Check if values supports output operations
    values_is_output_iterator = False
    try:
        output_deref = values.output_dereference
        if output_deref is not None:
            values_is_output_iterator = True
    except AttributeError:
        pass

    alloca_temp_for_index_type = make_alloca_intrinsic(indices.value_type)
    alloca_temp_for_value_state = make_alloca_intrinsic(values.state_type)
    get_value_state_field_ptr = make_struct_field_ptr_intrinsic(0)
    get_index_state_field_ptr = make_struct_field_ptr_intrinsic(1)

    # Create a local namespace to avoid polluting globals
    local_ns = {
        "alloca_temp_for_index_type": alloca_temp_for_index_type,
        "alloca_temp_for_value_state": alloca_temp_for_value_state,
        "get_value_state_field_ptr": get_value_state_field_ptr,
        "get_index_state_field_ptr": get_index_state_field_ptr,
    }

    # JIT compile value advance/dereference methods
    local_ns["value_advance"] = cuda.jit(values.advance, device=True)
    local_ns["value_input_dereference"] = cuda.jit(
        values.input_dereference, device=True
    )

    if values_is_output_iterator:
        local_ns["value_output_dereference"] = cuda.jit(
            values.output_dereference, device=True
        )

    # JIT compile index advance/dereference methods
    local_ns["index_advance"] = cuda.jit(indices.advance, device=True)
    local_ns["index_input_dereference"] = cuda.jit(
        indices.input_dereference, device=True
    )

    # Define the advance method
    advance_method_code = """
def permutation_advance(state, distance):
    # advance the index iterator
    index_state_ptr = get_index_state_field_ptr(state)
    index_advance(index_state_ptr, distance)
"""

    # Define the input dereference method
    input_dereference_method_code = """
def permutation_input_dereference(state, result):
    # dereference index to get the index value
    index_state_ptr = get_index_state_field_ptr(state)
    temp_index = alloca_temp_for_index_type()
    index_input_dereference(index_state_ptr, temp_index)

    # copy the value state (which always points to position 0)
    # and advance it by the index value
    value_state_ptr = get_value_state_field_ptr(state)
    temp_value_state = alloca_temp_for_value_state()
    temp_value_state[0] = value_state_ptr[0]
    value_advance(temp_value_state, temp_index[0])
    value_input_dereference(temp_value_state, result)
"""

    # Execute the method definitions in local namespace
    exec(advance_method_code, local_ns)
    exec(input_dereference_method_code, local_ns)

    advance_func = local_ns["permutation_advance"]
    input_dereference_func = local_ns["permutation_input_dereference"]
    output_dereference_func = None

    # Define the output dereference method if values supports it
    if values_is_output_iterator:
        output_dereference_method_code = """
def permutation_output_dereference(state, x):
    # dereference index to get the index value
    index_state_ptr = get_index_state_field_ptr(state)
    temp_index = alloca_temp_for_index_type()
    index_input_dereference(index_state_ptr, temp_index)

    # copy the value state (which always points to position 0)
    # and advance it by the index value
    value_state_ptr = get_value_state_field_ptr(state)
    temp_value_state = alloca_temp_for_value_state()
    temp_value_state[0] = value_state_ptr[0]
    value_advance(temp_value_state, temp_index[0])
    value_output_dereference(temp_value_state, x)
"""
        exec(output_dereference_method_code, local_ns)
        output_dereference_func = local_ns["permutation_output_dereference"]

    return advance_func, input_dereference_func, output_dereference_func


def make_permutation_iterator(values, indices):
    """
    Create a PermutationIterator that accesses values through an index mapping.

    The permutation iterator accesses elements from `values` using indices from `indices`,
    effectively computing values[indices[i]] at position i.

    Args:
        values: The values array or iterator to permute
        indices: The indices array or iterator specifying the permutation

    Returns:
        PermutationIterator: Iterator that yields permuted values
    """

    # Convert arrays to iterators if needed
    if hasattr(values, "__cuda_array_interface__"):
        values = pointer(values)
    elif not isinstance(values, IteratorBase):
        raise TypeError("values must be a device array or iterator")

    if hasattr(indices, "__cuda_array_interface__"):
        indices = pointer(indices)
    elif not isinstance(indices, IteratorBase):
        raise TypeError("indices must be an iterator or device array")

    # Get the cached advance and dereference methods
    advance_func, input_dereference_func, output_dereference_func = (
        _generate_advance_and_dereference_methods(values, indices)
    )

    # The cvalue and state for PermutationIterator are
    # structs composed of the cvalues and states of the
    # value and index iterators.
    from ..struct import gpu_struct

    class PermutationCValueStruct(ctypes.Structure):
        _fields_ = [
            ("value_state", values.cvalue.__class__),
            ("index_state", indices.cvalue.__class__),
        ]

    # Create state struct using gpu_struct
    PermutationState = gpu_struct(
        {"value_state": values.state_type, "index_state": indices.state_type},
        name="PermutationState",
    )
    state_type = PermutationState._type_descriptor

    cvalue = PermutationCValueStruct(values.cvalue, indices.cvalue)
    value_type = values.value_type

    class PermutationIterator(IteratorBase):
        def __init__(self, values_it, indices_it):
            self._values = values_it
            self._indices = indices_it
            kind = PermutationIteratorKind(
                value_type, self._values.kind, self._indices.kind
            )
            super().__init__(
                kind=kind,
                cvalue=cvalue,
                state_type=state_type,
                value_type=value_type,
            )

        @property
        def advance(self):
            return advance_func

        @property
        def input_dereference(self):
            return input_dereference_func

        @property
        def output_dereference(self):
            if output_dereference_func is None:
                raise AttributeError(
                    "PermutationIterator cannot be used as output iterator "
                    "when values iterator does not support output"
                )
            return output_dereference_func

        @property
        def children(self):
            return (self._values, self._indices)

        def _rebuild_value_type_from_children(self):
            self.value_type = self._values.value_type

    return PermutationIterator(values, indices)



# Copyright (c) 2024-2025, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from ._block_exchange import (
    BlockExchangeType,
    exchange,
)
from ._block_load_store import load, store
from ._block_merge_sort import merge_sort_keys
from ._block_radix_sort import (
    radix_sort_keys,
    radix_sort_keys_descending,
)
from ._block_reduce import reduce, sum
from ._block_scan import (
    exclusive_scan,
    exclusive_sum,
    inclusive_scan,
    inclusive_sum,
    scan,
)

__all__ = [
    "BlockExchangeType",
    "exchange",
    "exclusive_scan",
    "exclusive_sum",
    "inclusive_scan",
    "inclusive_sum",
    "load",
    "merge_sort_keys",
    "radix_sort_keys",
    "radix_sort_keys_descending",
    "reduce",
    "scan",
    "store",
    "sum",
]



# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from ._iterators import (
    CacheModifiedPointer as _CacheModifiedPointer,
)
from ._iterators import (
    ConstantIterator as _ConstantIterator,
)
from ._iterators import (
    CountingIterator as _CountingIterator,
)
from ._iterators import (
    DiscardIterator as _DiscardIterator,
)
from ._iterators import (
    make_reverse_iterator,
    make_transform_iterator,
)
from ._permutation_iterator import make_permutation_iterator
from ._zip_iterator import make_zip_iterator


def CacheModifiedInputIterator(device_array, modifier):
    """Random Access Cache Modified Iterator that wraps a native device pointer.

    Similar to https://nvidia.github.io/cccl/cub/api/classcub_1_1CacheModifiedInputIterator.html

    Currently the only supported modifier is "stream" (LOAD_CS).

    Example:
        The code snippet below demonstrates the usage of a ``CacheModifiedInputIterator``:

        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/iterator/cache_modified_iterator_basic.py
            :language: python
            :start-after: # example-begin


    Args:
        device_array: Array storing the input sequence of data items
        modifier: The PTX cache load modifier

    Returns:
        A ``CacheModifiedInputIterator`` object initialized with ``device_array``
    """
    from .. import types

    if modifier != "stream":
        raise NotImplementedError("Only stream modifier is supported")
    return _CacheModifiedPointer(
        device_array.__cuda_array_interface__["data"][0],
        types.from_numpy_dtype(device_array.dtype),
    )


def ConstantIterator(value):
    """Returns an Iterator representing a sequence of constant values.

    Similar to https://nvidia.github.io/cccl/thrust/api/classthrust_1_1constant__iterator.html

    Example:
        The code snippet below demonstrates the usage of a ``ConstantIterator``
        representing a sequence of constant values:

        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/iterator/constant_iterator_basic.py
            :language: python
            :start-after: # example-begin


    Args:
        value: The value of every item in the sequence

    Returns:
        A ``ConstantIterator`` object initialized to ``value``
    """
    return _ConstantIterator(value)


def CountingIterator(offset):
    """Returns an Iterator representing a sequence of incrementing values.

    Similar to https://nvidia.github.io/cccl/thrust/api/classthrust_1_1counting__iterator.html

    Example:
        The code snippet below demonstrates the usage of a ``CountingIterator``
        representing the sequence ``[10, 11, 12]``:

        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/iterator/counting_iterator_basic.py
            :language: python
            :start-after: # example-begin


    Args:
        offset: The initial value of the sequence

    Returns:
        A ``CountingIterator`` object initialized to ``offset``
    """
    return _CountingIterator(offset)


def DiscardIterator(reference_iterator=None):
    """Returns an Input or Output Iterator that discards all values written to it.

    Similar to https://nvidia.github.io/cccl/thrust/api/classthrust_1_1discard__iterator.html

    Args:
        reference_iterator: Optional iterator to use as a reference for value_type and state_type.
                          If not provided, defaults to uint8.

    Example:
        The code snippet below demonstrates the usage of a ``DiscardIterator`` to discard items in a unique by key operation:

        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/iterator/discard_iterator_basic.py
            :language: python
            :start-after: # example-begin


    Returns:
        A ``DiscardIterator`` object
    """
    return _DiscardIterator(reference_iterator)


def ReverseIterator(sequence):
    """Returns an Iterator over an array or another iterator in reverse.

    Similar to [std::reverse_iterator](https://en.cppreference.com/w/cpp/iterator/reverse_iterator).

    Examples:
        The code snippet below demonstrates the usage of a ``ReverseIterator`` as an input iterator:

        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/iterator/reverse_input_iterator.py
            :language: python
            :start-after: # example-begin

        The code snippet below demonstrates the usage of a ``ReverseIterator`` as an output iterator:

        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/iterator/reverse_output_iterator.py
            :language: python
            :start-after: # example-begin


    Args:
        sequence: The iterator or array to be reversed

    Returns:
        A ``ReverseIterator`` object
    """
    return make_reverse_iterator(sequence)


def TransformIterator(it, op):
    """An iterator that applies a user-defined unary function to the elements of an underlying iterator as they are read.

    Similar to [thrust::transform_iterator](https://nvidia.github.io/cccl/thrust/api/classthrust_1_1transform__iterator.html)

    Example:
        The code snippet below demonstrates the usage of a ``TransformIterator`` composed with a ``CountingIterator``
        to transform the input before performing a reduction.

        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/iterator/transform_iterator_basic.py
            :language: python
            :start-after: # example-begin
    Args:
        it: The underlying iterator
        op: The unary operation to be applied to values as they are read from ``it``

    Returns:
        A ``TransformIterator`` object to transform the items in ``it`` using ``op``
    """
    return make_transform_iterator(it, op, "input")


def TransformOutputIterator(it, op):
    """An iterator that applies a user-defined unary function to values before writing them to an underlying iterator.

    Similar to [thrust::transform_output_iterator](https://nvidia.github.io/cccl/thrust/api/classthrust_1_1transform__output__iterator.html).

    Example:
        The code snippet below demonstrates the usage of a ``TransformOutputIterator`` to transform the output
        of a reduction before writing to an output array.

        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/iterator/transform_output_iterator.py
            :language: python
            :start-after: # example-begin

    Args:
        it: The underlying iterator
        op: The operation to be applied to values before they are written to ``it``

    Returns:
        A ``TransformOutputIterator`` object that applies ``op`` to transform values before writing them to ``it``
    """
    return make_transform_iterator(it, op, "output")


def PermutationIterator(values, indices):
    """Returns an Iterator that accesses values through an index mapping.

    Similar to https://nvidia.github.io/cccl/thrust/api/classthrust_1_1permutation__iterator.html

    The permutation iterator accesses elements from the values collection using indices
    from the indices collection, effectively computing values[indices[i]] at position i.
    This is useful for gather/scatter operations and indirect array access patterns.

    Example:
        The code snippet below demonstrates the usage of a ``PermutationIterator``
        to access values in a permuted order:

        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/iterator/permutation_iterator_basic.py
            :language: python
            :start-after: # example-begin

    Args:
        values: The values array or iterator to be permuted
        indices: An iterator or device array providing the indices for permutation

    Returns:
        A ``PermutationIterator`` object that yields values[indices[i]] at position i
    """
    return make_permutation_iterator(values, indices)


def ZipIterator(*iterators):
    """Returns an Iterator representing a zipped sequence of values from N iterators.

    Similar to https://nvidia.github.io/cccl/thrust/api/classthrust_1_1zip__iterator.html

    The resulting iterator yields gpu_struct objects with fields corresponding to each input iterator.
    For 2 iterators, fields are named 'first' and 'second'. For N iterators, fields are indexed
    as field_0, field_1, ..., field_N-1.

    Example:
        The code snippet below demonstrates the usage of a ``ZipIterator``
        combining two device arrays:

        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/iterator/zip_iterator_elementwise.py
            :language: python
            :start-after: # example-begin

        ZipIterator can also be used with nested gpu_struct types:

        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/struct/nested_struct_zip_iterator.py
            :language: python
            :start-after: # example-begin

    Args:
        *iterators: Variable number of iterators to zip (at least 1)

    Returns:
        A ``ZipIterator`` object that yields combined values from all input iterators
    """
    return make_zip_iterator(*iterators)



# Copyright (c) 2024, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from ._warp_merge_sort import merge_sort_keys
from ._warp_reduce import reduce, sum
from ._warp_scan import exclusive_sum

__all__ = ["exclusive_sum", "reduce", "sum", "merge_sort_keys"]



# Copyright (c) 2024-2025, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from typing import Protocol

from .struct import _Struct


class DeviceArrayLike(Protocol):
    """
    Objects representing a device array, having a `.__cuda_array_interface__`
    attribute.
    """

    __cuda_array_interface__: dict


class StreamLike(Protocol):
    """
    Objects representing a stream, having a `.__cuda_stream__` attribute.
    """

    def __cuda_stream__(self) -> tuple[int, int]: ...


GpuStruct = _Struct



# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

"""Numba intrinsics and intrinsic factory functions for CUDA compute operations."""

import operator

import numba
from llvmlite import ir
from numba import types
from numba.core.extending import intrinsic, overload

_DEVICE_POINTER_SIZE = 8
_DEVICE_POINTER_BITWIDTH = _DEVICE_POINTER_SIZE * 8


def make_alloca_intrinsic(type_descriptor):
    """Create an intrinsic for allocating temporary storage of the given type.

    This factory creates a Numba intrinsic that allocates stack storage for
    a value of the specified type and returns a pointer to it.

    Args:
        type_descriptor: A TypeDescriptor (including PointerTypeDescriptor)

    Returns:
        An intrinsic function that, when called, returns a CPointer to
        allocated storage of the given type.

    Example:
        from cuda.compute import types
        alloca_temp = make_alloca_intrinsic(types.float32)
        # In a device function:
        temp_ptr = alloca_temp()  # Returns CPointer(float32)
    """
    from ._jit import type_descriptor_to_numba

    numba_type = type_descriptor_to_numba(type_descriptor)

    @intrinsic
    def alloca_temp(context):
        def codegen(context, builder, sig, args):
            temp_value_type = context.get_value_type(numba_type)
            temp_ptr = builder.alloca(temp_value_type)
            return temp_ptr

        return types.CPointer(numba_type)(), codegen

    return alloca_temp


def make_struct_field_ptr_intrinsic(field_idx: int):
    """Create an intrinsic for accessing a struct field by index.

    This factory creates a Numba intrinsic that takes a pointer to a struct
    and returns a pointer to the field at the specified index.

    Args:
        field_idx: The 0-based index of the field to access

    Returns:
        An intrinsic function that takes a struct pointer and returns
        a CPointer to the specified field.

    Example:
        get_first_field = make_struct_field_ptr_intrinsic(0)
        get_second_field = make_struct_field_ptr_intrinsic(1)
        # In a device function:
        field0_ptr = get_first_field(struct_ptr)  # Returns CPointer to field 0
    """
    from numba.core.datamodel.registry import default_manager

    @intrinsic
    def get_field_ptr(context, struct_ptr_type):
        def codegen(context, builder, sig, args):
            struct_ptr = args[0]
            field_ptr = builder.gep(
                struct_ptr,
                [
                    ir.Constant(ir.IntType(32), 0),
                    ir.Constant(ir.IntType(32), field_idx),
                ],
            )
            return field_ptr

        struct_model = default_manager.lookup(struct_ptr_type.dtype)
        field_type = struct_model._members[field_idx]
        return types.CPointer(field_type)(struct_ptr_type), codegen

    return get_field_ptr


def _sizeof_pointee(context, ptr):
    """Helper to get size of pointee type as an IR constant."""
    size = context.get_abi_sizeof(ptr.type.pointee)
    return ir.Constant(ir.IntType(_DEVICE_POINTER_BITWIDTH), size)


@intrinsic
def pointer_add_intrinsic(context, ptr, offset):
    """Intrinsic for pointer arithmetic (ptr + offset)."""

    def codegen(context, builder, sig, args):
        ptr, index = args
        base = builder.ptrtoint(ptr, ir.IntType(_DEVICE_POINTER_BITWIDTH))
        sizeof = _sizeof_pointee(context, ptr)
        # Cast index to match sizeof type if needed
        if index.type != sizeof.type:
            index = (
                builder.sext(index, sizeof.type)
                if index.type.width < sizeof.type.width
                else builder.trunc(index, sizeof.type)
            )
        offset = builder.mul(index, sizeof)
        result = builder.add(base, offset)
        return builder.inttoptr(result, ptr.type)

    return ptr(ptr, offset), codegen


@overload(operator.add)
def pointer_add(ptr, offset):
    """Overload operator.add for CPointer + Integer."""
    if not isinstance(ptr, numba.types.CPointer) or not isinstance(
        offset, numba.types.Integer
    ):
        return

    def impl(ptr, offset):
        return pointer_add_intrinsic(ptr, offset)

    return impl


@intrinsic
def load_cs(typingctx, base):
    """Load with cache streaming hint (ld.global.cs).

    Corresponding to `LOAD_CS` here:
    https://nvidia.github.io/cccl/cub/api/classcub_1_1CacheModifiedInputIterator.html
    """

    def codegen(context, builder, sig, args):
        rt = context.get_value_type(sig.return_type)
        if rt is None:
            raise RuntimeError(f"Unsupported return type: {type(sig.return_type)}")
        ftype = ir.FunctionType(rt, [rt.as_pointer()])
        bw = sig.return_type.bitwidth
        asm_txt = f"ld.global.cs.b{bw} $0, [$1];"
        if bw < 64:
            constraint = "=r, l"
        else:
            constraint = "=l, l"
        asm_ir = ir.InlineAsm(ftype, asm_txt, constraint)
        return builder.call(asm_ir, args)

    return base.dtype(base), codegen


__all__ = [
    "load_cs",
    "make_alloca_intrinsic",
    "make_struct_field_ptr_intrinsic",
    "pointer_add_intrinsic",
]



# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

import ctypes

from numba import cuda

from .. import types
from .._caching import cache_with_registered_key_functions
from ._iterators import (
    IteratorBase,
    IteratorKind,
    pointer,
)


class ZipIteratorKind(IteratorKind):
    def __init__(
        self,
        value_type: types.TypeDescriptor,
        iterators_kinds: tuple[IteratorKind, ...],
    ):
        self.iterators_kinds = iterators_kinds
        self.value_type = value_type

    def __repr__(self):
        return f"ZipIteratorKind({self.value_type}, {self.iterators_kinds})"

    def __eq__(self, other):
        return (
            type(self) is type(other)
            and self.value_type == other.value_type
            and self.iterators_kinds == other.iterators_kinds
        )

    def __hash__(self):
        return hash((type(self), self.value_type, self.iterators_kinds))


def _get_zip_iterator_metadata(iterators):
    from ..struct import gpu_struct

    # Create ctypes struct for cvalue
    fields = [(f"iter_{i}", it.cvalue.__class__) for i, it in enumerate(iterators)]
    ZipCValueStruct = type("ZipCValueStruct", (ctypes.Structure,), {"_fields_": fields})

    # Create state struct using gpu_struct
    state_fields = {f"state_{i}": it.state_type for i, it in enumerate(iterators)}
    ZipState = gpu_struct(state_fields, name="ZipState")
    state_type = ZipState._type_descriptor

    # Create value struct using gpu_struct
    value_fields = {f"value_{i}": it.value_type for i, it in enumerate(iterators)}
    ZipValue = gpu_struct(value_fields, name="ZipValue")
    value_type = ZipValue._type_descriptor

    cvalue = ZipCValueStruct(
        **{f"iter_{i}": it.cvalue for i, it in enumerate(iterators)}
    )
    return cvalue, state_type, value_type


# Automatic: iterators tuple → (kind, kind, ...) or (dtype, dtype, ...)
@cache_with_registered_key_functions
def _get_advance_and_dereference_functions(iterators):
    # Generate the advance and dereference functions for the zip iterator
    # composed of the input iterators
    # Put simply, the advance method invokes the advance method of each input
    # iterator, and the dereference method invokes the dereference method of each
    # input iterator.

    from .._intrinsics import make_struct_field_ptr_intrinsic

    n_iterators = len(iterators)

    # Create a local namespace for this zip iterator to avoid polluting globals
    # and prevent name collisions when nesting zip iterators
    local_ns = {}

    # Within the advance and dereference methods of this iterator, we need a way
    # to get pointers to the fields of the state struct (advance and dereference),
    # and the value struct (dereference). This needs `n` intrinsics, one
    # for each field. Create them using the factory:
    for field_idx in range(n_iterators):
        local_ns[f"get_field_ptr_{field_idx}"] = make_struct_field_ptr_intrinsic(
            field_idx
        )

    # Now we can define the advance and dereference methods of this iterator,
    # which also need to be defined dynamically because they will use the
    # intrinsics defined above.
    for i, it in enumerate(iterators):
        local_ns[f"advance_{i}"] = cuda.jit(it.advance, device=True)
        local_ns[f"input_dereference_{i}"] = cuda.jit(it.input_dereference, device=True)
        # Also compile output_dereference if available
        try:
            output_deref = it.output_dereference
            if output_deref is not None:
                local_ns[f"output_dereference_{i}"] = cuda.jit(
                    output_deref, device=True
                )
        except AttributeError:
            # Iterator doesn't support output operations
            pass

    # Generate the advance method, which advances each input iterator:
    advance_lines = []  # lines of code for the advance method
    input_dereference_lines = []  # lines of code for input dereference method
    output_dereference_lines = []  # lines of code for output dereference method

    # Check if all iterators support output operations
    def supports_output(it):
        try:
            return it.output_dereference is not None
        except AttributeError:
            return False

    all_support_output = all(supports_output(it) for it in iterators)

    for i in range(n_iterators):
        advance_lines.append(
            f"    state_ptr_{i} = get_field_ptr_{i}(state)\n"
            f"    advance_{i}(state_ptr_{i}, distance)"
        )
        input_dereference_lines.append(
            f"    state_ptr_{i} = get_field_ptr_{i}(state)\n"
            f"    result_ptr_{i} = get_field_ptr_{i}(result)\n"
            f"    input_dereference_{i}(state_ptr_{i}, result_ptr_{i})"
        )
        if all_support_output:
            output_dereference_lines.append(
                f"    state_ptr_{i} = get_field_ptr_{i}(state)\n"
                f"    output_dereference_{i}(state_ptr_{i}, x.value_{i})"
            )

    advance_method_code = f"""
def input_advance(state, distance):
    # Advance each iterator using dynamically created field pointer functions
{chr(10).join(advance_lines)}
"""  # chr(10) is '\n'

    input_dereference_method_code = f"""
def input_dereference(state, result):
    # Dereference each iterator using dynamically created field pointer functions
{chr(10).join(input_dereference_lines)}
"""  # chr(10) is '\n'

    # Execute the method codes in local namespace:
    exec(advance_method_code, local_ns)
    exec(input_dereference_method_code, local_ns)

    advance_func = local_ns["input_advance"]
    input_dereference_func = local_ns["input_dereference"]

    # Generate output_dereference if all iterators support it
    output_dereference_func = None
    if all_support_output:
        output_dereference_method_code = f"""
def output_dereference(state, x):
    # Write to each iterator using dynamically created field pointer functions
{chr(10).join(output_dereference_lines)}
"""
        exec(output_dereference_method_code, local_ns)
        output_dereference_func = local_ns["output_dereference"]

    return advance_func, input_dereference_func, output_dereference_func


def make_zip_iterator(*iterators):
    """
    Create a ZipIterator that combines N iterators.

    Args:
        *iterators: Variable number of iterators or device arrays

    Returns:
        ZipIterator: Iterator that combines all input iterators
    """
    if len(iterators) < 1:
        raise ValueError("At least 1 iterator is required")

    # Convert arrays to iterators if needed
    processed_iterators = []
    for it in iterators:
        if hasattr(it, "__cuda_array_interface__"):
            it = pointer(it)
        processed_iterators.append(it)

    # Validate all are iterators
    for i, it in enumerate(processed_iterators):
        if not isinstance(it, IteratorBase):
            raise TypeError(f"Argument {i} must be an iterator or device array")

    cvalue, state_type, value_type = _get_zip_iterator_metadata(processed_iterators)

    advance_func, input_dereference_func, output_dereference_func = (
        _get_advance_and_dereference_functions(processed_iterators)
    )

    # Check if all underlying iterators support output
    def supports_output(it):
        try:
            return it.output_dereference is not None
        except AttributeError:
            return False

    all_support_output = all(supports_output(it) for it in processed_iterators)

    class ZipIterator(IteratorBase):
        def __init__(self, iterators_list):
            self._iterators = iterators_list
            self._n_iterators = len(iterators_list)

            kinds = [it.kind for it in iterators_list]
            kind = ZipIteratorKind(value_type, tuple(kinds))
            super().__init__(
                kind=kind,
                cvalue=cvalue,
                state_type=state_type,
                value_type=value_type,
            )

        @property
        def host_advance(self):
            return advance_func

        @property
        def advance(self):
            return advance_func

        @property
        def input_dereference(self):
            return input_dereference_func

        @property
        def output_dereference(self):
            if not all_support_output:
                raise AttributeError(
                    "ZipIterator cannot be used as output iterator when not all "
                    "underlying iterators support output"
                )
            return output_dereference_func

        @property
        def children(self):
            return tuple(self._iterators)

        def _rebuild_value_type_from_children(self):
            from ..struct import gpu_struct

            if not self._iterators:
                raise ValueError("Zip iterator has no children")
            value_fields = {
                f"value_{i}": child.value_type
                for i, child in enumerate(self._iterators)
            }
            self.value_type = gpu_struct(value_fields, name="ZipValue")._type_descriptor

    return ZipIterator(processed_iterators)



# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from typing import Callable

import numpy as np

from .. import _bindings
from .. import _cccl_interop as cccl
from .._caching import cache_with_registered_key_functions
from .._cccl_interop import (
    call_build,
    get_value_type,
    set_cccl_iterator_state,
    to_cccl_value_state,
)
from .._utils.protocols import (
    get_data_pointer,
    validate_and_get_stream,
)
from .._utils.temp_storage_buffer import TempStorageBuffer
from ..iterators._iterators import IteratorBase
from ..op import OpAdapter, OpKind, make_op_adapter
from ..typing import DeviceArrayLike, GpuStruct


class _SegmentedReduce:
    __slots__ = [
        "build_result",
        "d_in_cccl",
        "d_out_cccl",
        "start_offsets_in_cccl",
        "end_offsets_in_cccl",
        "h_init_cccl",
        "op_cccl",
    ]

    def __init__(
        self,
        d_in: DeviceArrayLike | IteratorBase,
        d_out: DeviceArrayLike | IteratorBase,
        start_offsets_in: DeviceArrayLike | IteratorBase,
        end_offsets_in: DeviceArrayLike | IteratorBase,
        op: OpAdapter,
        h_init: np.ndarray | GpuStruct,
    ):
        self.d_in_cccl = cccl.to_cccl_input_iter(d_in)
        self.d_out_cccl = cccl.to_cccl_output_iter(d_out)
        self.start_offsets_in_cccl = cccl.to_cccl_input_iter(start_offsets_in)
        self.end_offsets_in_cccl = cccl.to_cccl_input_iter(end_offsets_in)
        self.h_init_cccl = cccl.to_cccl_value(h_init)

        # Compile the op with value types
        value_type = get_value_type(h_init)

        self.op_cccl = op.compile((value_type, value_type), value_type)

        self.build_result = call_build(
            _bindings.DeviceSegmentedReduceBuildResult,
            self.d_in_cccl,
            self.d_out_cccl,
            self.start_offsets_in_cccl,
            self.end_offsets_in_cccl,
            self.op_cccl,
            self.h_init_cccl,
        )

    def __call__(
        self,
        temp_storage,
        d_in,
        d_out,
        op: Callable | OpAdapter,
        num_segments: int,
        start_offsets_in,
        end_offsets_in,
        h_init,
        stream=None,
    ):
        if num_segments > np.iinfo(np.int32).max:
            raise RuntimeError(
                "Segmented sort does not currently support more than 2^31-1 segments."
            )
        set_cccl_iterator_state(self.d_in_cccl, d_in)
        set_cccl_iterator_state(self.d_out_cccl, d_out)
        set_cccl_iterator_state(self.start_offsets_in_cccl, start_offsets_in)
        set_cccl_iterator_state(self.end_offsets_in_cccl, end_offsets_in)

        # Update op state for stateful ops
        op_adapter = make_op_adapter(op)
        op_adapter.update_op_state(self.op_cccl)

        self.h_init_cccl.state = to_cccl_value_state(h_init)

        stream_handle = validate_and_get_stream(stream)

        if temp_storage is None:
            temp_storage_bytes = 0
            d_temp_storage = 0
        else:
            temp_storage_bytes = temp_storage.nbytes
            d_temp_storage = get_data_pointer(temp_storage)

        temp_storage_bytes = self.build_result.compute(
            d_temp_storage,
            temp_storage_bytes,
            self.d_in_cccl,
            self.d_out_cccl,
            num_segments,
            self.start_offsets_in_cccl,
            self.end_offsets_in_cccl,
            self.op_cccl,
            self.h_init_cccl,
            stream_handle,
        )
        return temp_storage_bytes


@cache_with_registered_key_functions
def make_segmented_reduce(
    d_in: DeviceArrayLike | IteratorBase,
    d_out: DeviceArrayLike | IteratorBase,
    start_offsets_in: DeviceArrayLike | IteratorBase,
    end_offsets_in: DeviceArrayLike | IteratorBase,
    op: Callable | OpKind,
    h_init: np.ndarray | GpuStruct,
):
    """Computes a device-wide segmented reduction using the specified binary ``op`` and initial value ``init``.

    Example:
        Below, ``make_segmented_reduce`` is used to create a segmented reduction object that can be reused.

        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/segmented/segmented_reduce_object.py
            :language: python
            :start-after: # example-begin


    Args:
        d_in: Device array or iterator containing the input sequence of data items
        d_out: Device array that will store the result of the reduction
        start_offsets_in: Device array or iterator containing offsets to start of segments
        end_offsets_in: Device array or iterator containing offsets to end of segments
        op: Callable or OpKind representing the binary operator to apply
        init: Numpy array storing initial value of the reduction

    Returns:
        A callable object that can be used to perform the reduction
    """
    op_adapter = make_op_adapter(op)
    return _SegmentedReduce(
        d_in, d_out, start_offsets_in, end_offsets_in, op_adapter, h_init
    )


def segmented_reduce(
    d_in: DeviceArrayLike | IteratorBase,
    d_out: DeviceArrayLike | IteratorBase,
    start_offsets_in: DeviceArrayLike | IteratorBase,
    end_offsets_in: DeviceArrayLike | IteratorBase,
    op: Callable | OpKind,
    h_init: np.ndarray | GpuStruct,
    num_segments: int,
    stream=None,
):
    """
    Performs device-wide segmented reduction.

    This function automatically handles temporary storage allocation and execution.

    Example:
        Below, ``segmented_reduce`` is used to compute the minimum value of segments in a sequence of integers.

        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/segmented/segmented_reduce_basic.py
            :language: python
            :start-after: # example-begin


    Args:
        d_in: Device array or iterator containing the input sequence of data items
        d_out: Device array to store the result of the reduction for each segment
        start_offsets_in: Device array or iterator containing the sequence of beginning offsets
        end_offsets_in: Device array or iterator containing the sequence of ending offsets
        op: Binary reduction operator
        h_init: Initial value for the reduction
        num_segments: Number of segments to reduce
        stream: CUDA stream for the operation (optional)
    """
    reducer = make_segmented_reduce(
        d_in, d_out, start_offsets_in, end_offsets_in, op, h_init
    )
    tmp_storage_bytes = reducer(
        None,
        d_in,
        d_out,
        op,
        num_segments,
        start_offsets_in,
        end_offsets_in,
        h_init,
        stream,
    )
    tmp_storage = TempStorageBuffer(tmp_storage_bytes, stream)
    reducer(
        tmp_storage,
        d_in,
        d_out,
        op,
        num_segments,
        start_offsets_in,
        end_offsets_in,
        h_init,
        stream,
    )



# Copyright (c) 2026, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

import ast
import functools
import inspect
import operator
import struct
import textwrap
import uuid
from types import new_class
from typing import Callable, Hashable, List, Tuple, get_type_hints

import numba
import numba.cuda
import numba.np.numpy_support
import numba.types
import numpy as np
from numba import types
from numba.core import cgutils
from numba.core.datamodel import models
from numba.core.extending import (
    as_numba_type,
    make_attribute_wrapper,
    overload,
    register_model,
    typeof_impl,
)
from numba.core.typeconv import Conversion
from numba.core.typing import signature as nb_signature
from numba.core.typing.templates import ConcreteTemplate
from numba.cuda.cudadecl import registry as cuda_registry
from numba.extending import lower_builtin, lower_cast

from . import types as cccl_types
from ._bindings import Op, OpKind
from ._caching import CachableFunction, cache_with_registered_key_functions
from ._odr_helpers import create_stateful_op_void_ptr_wrapper
from ._utils import sanitize_identifier
from ._utils.protocols import (
    get_data_pointer,
    get_dtype,
    is_contiguous,
    is_device_array,
)
from .op import OpAdapter
from .typing import DeviceArrayLike

# -----------------------------------------------------------------------------
# Struct registration and casting
# -----------------------------------------------------------------------------


# Base class for all struct types, used for struct-to-struct cast matching.
class _StructBase(numba.types.Type):
    """Base class for all CCCL GPU struct types."""

    _field_spec: dict  # Mapping of field names to Numba types


# The struct registration logic is isolated here to avoid polluting other
# modules with Numba-specific type plumbing.
@functools.lru_cache(maxsize=256)
def _make_struct_type(struct_class_or_name, field_names, field_types):
    """
    Core factory function that uses the Numba extension machinery to
    create a struct type with pass-by-value semantics.

    This function is called lazily from _register_struct_with_numba.

    Args:
        struct_class_or_name: Either an existing struct class (from gpu_struct) or a string name
        field_names: Tuple of field names
        field_types: Tuple of TypeDescriptors
    """
    from .struct import _is_struct_type, _Struct

    # If we're given an existing class, use it; otherwise create a new one
    if isinstance(struct_class_or_name, type):
        struct_class = struct_class_or_name
    else:
        struct_class = new_class(struct_class_or_name, bases=(_Struct,))

    # Convert TypeDescriptors to Numba types
    numba_field_types = [type_descriptor_to_numba(ft) for ft in field_types]

    raw_field_spec = dict(zip(field_names, numba_field_types))
    assert all(
        _is_struct_type(tp) or isinstance(tp, types.Type)
        for tp in raw_field_spec.values()
    )

    field_spec = {
        name: as_numba_type(typ) if _is_struct_type(typ) else typ
        for name, typ in raw_field_spec.items()
    }

    # Only set _field_spec if it doesn't already exist (for new classes created here)
    # For classes from gpu_struct, they already have _field_spec with TypeDescriptors
    if not hasattr(struct_class, "_field_spec"):
        struct_class._field_spec = raw_field_spec

    class StructType(_StructBase):
        def __init__(self):
            super().__init__(name=struct_class.__name__)
            # Store field_spec for struct-to-struct conversion checks
            self._field_spec = field_spec

        def can_convert_from(self, typingctx, other):
            if isinstance(other, types.UniTuple):
                tuple_size = other.count
                if tuple_size == len(field_types):
                    return Conversion.safe

            elif isinstance(other, types.Tuple):
                tuple_size = len(other.types)
                if tuple_size == len(field_types):
                    all_compatible = all(
                        typingctx.can_convert(src_type, tgt_type) is not None
                        for src_type, tgt_type in zip(other.types, field_spec.values())
                    )
                    if all_compatible:
                        return Conversion.safe

            # Allow conversion from another StructType with identical field layout
            elif hasattr(other, "_field_spec"):
                other_field_spec = other._field_spec
                # Check if field count matches
                if len(other_field_spec) == len(field_spec):
                    # Check if all field types are compatible (by position, not name)
                    other_field_types = list(other_field_spec.values())
                    self_field_types = list(field_spec.values())
                    all_compatible = all(
                        typingctx.can_convert(src_type, tgt_type) is not None
                        for src_type, tgt_type in zip(
                            other_field_types, self_field_types
                        )
                    )
                    if all_compatible:
                        return Conversion.safe

            return None

    numba_type = StructType()
    numba_type.python_type = struct_class

    as_numba_type.register(struct_class, numba_type)

    @typeof_impl.register(struct_class)
    def typeof_struct(val, c):
        return numba_type  # Must return the SAME instance, not a new StructType()

    @register_model(StructType)
    class StructModel(models.StructModel):
        def __init__(self, dmm, fe_type):
            members = [(name, typ) for name, typ in field_spec.items()]
            super().__init__(dmm, fe_type, members)

    for field_name in field_spec:
        make_attribute_wrapper(StructType, field_name, field_name)

    field_names_list = list(field_spec.keys())

    @overload(operator.getitem)
    def struct_getitem(struct_val, idx):
        if not isinstance(struct_val, StructType):
            return

        if isinstance(idx, (types.IntegerLiteral)):
            idx_val = getattr(idx, "literal_value", getattr(idx, "value", None))

            if idx_val is None or not (0 <= idx_val < len(field_names_list)):

                def error_impl(struct_val, idx):
                    raise IndexError(
                        f"Index out of range for struct with {len(field_names_list)} fields"
                    )

                return error_impl

            field_name = field_names_list[idx_val]
            exec(
                f"def impl(struct_val, idx): return struct_val.{field_name}",
                namespace := {},
            )
            return namespace["impl"]

        conditions = "\n".join(
            f"    {'if' if i == 0 else 'elif'} idx == {i}: return struct_val.{name}"
            for i, name in enumerate(field_names_list)
        )
        exec(
            f"def impl(struct_val, idx):\n{conditions}\n    else: raise IndexError('Index out of range')",
            namespace := {},
        )
        return namespace["impl"]

    @cuda_registry.register
    class StructConstructor(ConcreteTemplate):
        key = struct_class
        cases = [nb_signature(numba_type, *list(field_spec.values()))]

    cuda_registry.register_global(struct_class, numba.types.Function(StructConstructor))

    def struct_constructor(context, builder, sig, args):
        ty = sig.return_type
        retval = cgutils.create_struct_proxy(ty)(context, builder)
        for field_name, val in zip(field_spec.keys(), args):
            setattr(retval, field_name, val)
        return retval._getvalue()

    lower_builtin(struct_class, *list(field_spec.values()))(struct_constructor)

    @lower_cast(types.BaseTuple, StructType)
    def tuple_to_struct_cast(context, builder, fromty, toty, val):
        if isinstance(fromty, types.UniTuple):
            tuple_size = fromty.count
            element_types = [fromty.dtype] * tuple_size
        elif isinstance(fromty, types.Tuple):
            tuple_size = len(fromty.types)
            element_types = list(fromty.types)
        else:
            tuple_size = len(field_spec)
            element_types = list(field_spec.values())

        if tuple_size != len(field_spec):
            raise ValueError(
                f"Cannot cast tuple of size {tuple_size} to {struct_class.__name__} "
                f"with {len(field_types)} fields"
            )

        retval = cgutils.create_struct_proxy(toty)(context, builder)

        for i, (field_name, target_type) in enumerate(field_spec.items()):
            element = builder.extract_value(val, i)

            source_type = element_types[i]
            if source_type != target_type:
                element = context.cast(builder, element, source_type, target_type)

            setattr(retval, field_name, element)

        return retval._getvalue()

    @lower_cast(types.Tuple, StructType)
    @lower_cast(types.UniTuple, StructType)
    def cast_tuple_to_struct(context, builder, fromty, toty, val):
        if isinstance(fromty, types.UniTuple):
            if fromty.count != len(field_spec):
                return None
            tuple_types = [fromty.dtype] * fromty.count
        else:
            if len(fromty.types) != len(field_spec):
                return None
            tuple_types = list(fromty.types)

        struct_val = cgutils.create_struct_proxy(toty)(context, builder)
        for i, (field_name, field_type) in enumerate(field_spec.items()):
            elem = builder.extract_value(val, i)
            elem = context.cast(builder, elem, tuple_types[i], field_type)
            setattr(struct_val, field_name, elem)

        return struct_val._getvalue()

    @lower_cast(_StructBase, StructType)
    def cast_struct_to_struct(context, builder, fromty, toty, val):
        """Cast from one CCCL struct type to another with identical layout."""
        # Get field specs from both types
        from_field_spec = fromty._field_spec
        to_field_spec = toty._field_spec

        if len(from_field_spec) != len(to_field_spec):
            return None

        from_field_types = list(from_field_spec.values())
        from_field_names = list(from_field_spec.keys())
        to_field_types = list(to_field_spec.values())
        to_field_names = list(to_field_spec.keys())

        # Create struct proxy for source value
        from_struct = cgutils.create_struct_proxy(fromty)(context, builder, value=val)

        # Create struct proxy for target value
        to_struct = cgutils.create_struct_proxy(toty)(context, builder)

        # Copy and cast each field by position
        for i, (to_name, to_type) in enumerate(zip(to_field_names, to_field_types)):
            from_name = from_field_names[i]
            from_type = from_field_types[i]

            # Get the field value from source struct
            elem = getattr(from_struct, from_name)

            # Cast if types differ
            if from_type != to_type:
                elem = context.cast(builder, elem, from_type, to_type)

            # Set the field in target struct
            setattr(to_struct, to_name, elem)

        return to_struct._getvalue()

    return struct_class


@functools.lru_cache(maxsize=256)
def _register_struct_with_numba(struct_class):
    field_spec = struct_class._type_descriptor.fields

    registered_class = _make_struct_type(
        struct_class,
        tuple(field_spec.keys()),
        tuple(field_spec.values()),
    )

    return as_numba_type(registered_class)


# -----------------------------------------------------------------------------
# TypeDescriptor <-> Numba conversions
# -----------------------------------------------------------------------------


@functools.cache
def type_descriptor_to_numba(td):
    """
    Convert a TypeDescriptor to a Numba type.

    Handles:
    - PointerTypeDescriptor: creates CPointer to the pointee's numba type
    - StructTypeDescriptor: registers a struct class for the layout
    - POD TypeDescriptor: uses numba.from_dtype
    - Numba types: pass through
    """
    from . import types as cccl_types

    # Pass through if already a Numba type
    if isinstance(td, numba.types.Type):
        return td

    # Handle PointerTypeDescriptor (must check before TypeDescriptor since it's a subclass)
    if isinstance(td, cccl_types.PointerTypeDescriptor):
        return types.CPointer(type_descriptor_to_numba(td.pointee))

    # Handle TypeDescriptor (includes StructTypeDescriptor)
    if isinstance(td, cccl_types.TypeDescriptor):
        return _convert_type_descriptor_to_numba(td)

    raise TypeError(f"Expected TypeDescriptor or numba type, got {type(td)}")


def _convert_type_descriptor_to_numba(td):
    """Internal helper to convert TypeDescriptor to Numba type."""
    from . import types as cccl_types

    # For struct types
    if isinstance(td, cccl_types.StructTypeDescriptor):
        from .struct import (
            _get_struct_record_dtype,
            _get_struct_type_descriptor,
            _Struct,
        )

        layout_key = tuple(td.layout_key())
        struct_class = new_class(td.name, bases=(_Struct,))
        struct_class._field_spec = dict(layout_key)
        struct_class._type_descriptor = _get_struct_type_descriptor(struct_class)
        struct_class.dtype = _get_struct_record_dtype(struct_class)
        try:
            return as_numba_type(struct_class)
        except numba.core.errors.NumbaError:
            return _register_struct_with_numba(struct_class)

    # For POD types
    return numba.from_dtype(td.dtype)


def _is_gpu_struct_class(obj):
    """Check if an object is a gpu_struct class (not instance)."""
    return (
        isinstance(obj, type)
        and hasattr(obj, "_type_descriptor")
        and hasattr(obj, "_field_spec")
    )


def _iter_function_objects(py_func):
    if hasattr(py_func, "__globals__"):
        yield from py_func.__globals__.values()

    if py_func.__closure__ is not None:
        for cell in py_func.__closure__:
            try:
                yield cell.cell_contents
            except ValueError:
                pass


def _ensure_function_structs_registered(py_func):
    """
    Scan a function's globals and closure for gpu_struct classes and ensure
    they're registered with Numba before compilation.
    """

    def _register_if_needed(struct_class):
        try:
            return as_numba_type(struct_class)
        except numba.core.errors.NumbaError:
            return _register_struct_with_numba(struct_class)

    for value in _iter_function_objects(py_func):
        if _is_gpu_struct_class(value):
            _register_if_needed(value)


def _numba_type_to_type_descriptor(numba_type):
    """Convert a Numba type to a TypeDescriptor (internal helper)."""
    from . import types as cccl_types
    from .struct import _is_struct_type

    # Already a TypeDescriptor
    if isinstance(numba_type, cccl_types.TypeDescriptor):
        return numba_type

    # Custom StructType with an associated gpu_struct Python class
    if hasattr(numba_type, "python_type") and _is_struct_type(numba_type.python_type):
        return numba_type.python_type._type_descriptor

    # POD type - convert via numpy dtype
    dtype = numba.np.numpy_support.as_dtype(numba_type)
    return cccl_types.from_numpy_dtype(dtype)


def _annotation_to_type_descriptor(annotation):
    """
    Convert a type annotation to a TypeDescriptor.

    Handles:
    - TypeDescriptor: returns as-is
    - gpu_struct classes: returns their _type_descriptor
    - numpy dtypes/types: converts via from_numpy_dtype
    """
    from . import types as cccl_types
    from .struct import _is_struct_type

    if isinstance(annotation, cccl_types.TypeDescriptor):
        return annotation

    if _is_struct_type(annotation):
        return annotation._type_descriptor  # type: ignore[union-attr]

    # numpy dtype or type
    return cccl_types.from_numpy_dtype(np.dtype(annotation))


# -----------------------------------------------------------------------------
# Signature inference
# -----------------------------------------------------------------------------


def _infer_signature(py_func, input_types=None):
    """
    Infer the signature of a function as TypeDescriptors.

    If annotations are provided, uses those directly.
    Otherwise, compiles the function to infer types.

    Args:
        py_func: The Python function to analyze
        input_types: Optional tuple of TypeDescriptors for the inputs.
                     Used for inference when annotations are missing.

    Returns:
        Tuple of (input_type_descriptors, output_type_descriptor)
        where input_type_descriptors is a tuple of TypeDescriptor
    """
    try:
        annotations = get_type_hints(py_func)
    except Exception:
        annotations = py_func.__annotations__
    spec = inspect.getfullargspec(py_func)
    arg_names = list(spec.args)

    input_tds = []
    has_all_input_annotations = True

    # Try to get input types from annotations
    for name in arg_names:
        if name in annotations:
            input_tds.append(_annotation_to_type_descriptor(annotations[name]))
        else:
            has_all_input_annotations = False
            break

    # Try to get output type from return annotation
    output_td = None
    if "return" in annotations:
        output_td = _annotation_to_type_descriptor(annotations["return"])

    # If we have all annotations, we're done
    if has_all_input_annotations and output_td is not None:
        return tuple(input_tds), output_td

    # Need to infer by compiling
    if input_types is None:
        if not has_all_input_annotations:
            raise ValueError(
                "Function must have type annotations for all arguments, "
                "or input_types must be provided"
            )
        input_tds_for_compile = input_tds
    else:
        # Use provided input types (TypeDescriptors)
        input_tds_for_compile = input_types

    # Convert TypeDescriptors to Numba types for compilation
    input_numba_types = tuple(
        type_descriptor_to_numba(td) for td in input_tds_for_compile
    )

    # Ensure any gpu_struct classes referenced in the function are registered
    _ensure_function_structs_registered(py_func)

    # Compile to infer return type
    from ._utils import sanitize_identifier

    sanitized_name = sanitize_identifier(py_func.__name__)
    unique_suffix = hex(id(py_func))[2:]
    abi_name = f"{sanitized_name}_{unique_suffix}"
    _, return_type = numba.cuda.compile(
        py_func, input_numba_types, abi_info={"abi_name": abi_name}
    )
    output_td = _numba_type_to_type_descriptor(return_type)

    # Use provided input types or convert from numba types
    if input_types is not None:
        input_tds = list(input_types)
    elif not has_all_input_annotations:
        input_tds = [_numba_type_to_type_descriptor(t) for t in input_numba_types]

    return tuple(input_tds), output_td


# -----------------------------------------------------------------------------
# Iterator compilation
# -----------------------------------------------------------------------------


@functools.lru_cache(maxsize=256)
def _cached_compile(func, sig, abi_name=None, **kwargs):
    """Cached wrapper around numba.cuda.compile."""
    return numba.cuda.compile(func, sig, abi_info={"abi_name": abi_name}, **kwargs)


def _get_abi_suffix():
    """Generate a unique ABI suffix."""
    return uuid.uuid4().hex


def _resolve_iterator_value_types(it):
    # transform iterators sometimes need help figuring their input or
    # output types (depending on whether it's an input or output
    # iterator). This requires inspecting type annotations, or
    # using numba's type inference. This function takes an iterator
    # (possibly a compound iterator like ZipIterator) and traverses
    # its children recursively, finding any TransformIterators
    # and setting their value types. At the end, it calls
    # it._rebuild_value_type_from_children() which propagates
    # the updated value types back up to the "parent" iterators.
    from .iterators._iterators import TransformIteratorKind

    children = getattr(it, "children", ())
    for child in children:
        _resolve_iterator_value_types(child)

    kind = getattr(it, "kind", None)
    if isinstance(kind, TransformIteratorKind):
        op_func = kind.op._func
        _ensure_function_structs_registered(op_func)
        if kind.io_kind == "input":
            _, output_td = _infer_signature(op_func, (it.value_type,))
            it.value_type = output_td
        else:
            input_tds, _ = _infer_signature(op_func)
            it.value_type = input_tds[0]

    rebuild_value_type = getattr(it, "_rebuild_value_type_from_children", None)
    if rebuild_value_type is not None:
        rebuild_value_type()


def compile_iterator(it, io_kind: str):
    """
    Compile an iterator into a CCCL Iterator binding object.

    Args:
        it: The iterator to compile (an IteratorBase instance)
        io_kind: Either "input" or "output"

    Returns:
        An Iterator binding object ready for use with CCCL algorithms
    """
    from ._bindings import Iterator, IteratorKind, Op, OpKind
    from ._odr_helpers import (
        create_advance_void_ptr_wrapper,
        create_input_dereference_void_ptr_wrapper,
        create_output_dereference_void_ptr_wrapper,
    )

    _resolve_iterator_value_types(it)

    # Convert TypeDescriptors to Numba types for compilation
    numba_state_type = type_descriptor_to_numba(it.state_type)
    state_ptr_type = types.CPointer(numba_state_type)
    numba_value_type = type_descriptor_to_numba(it.value_type)

    # Validate state size using TypeDescriptor info
    # (PointerTypeDescriptor.info() returns 8 bytes, TypeDescriptor.info() returns dtype size)
    state_info = it.state_type.info()
    iterator_state = memoryview(it.state)
    if iterator_state.nbytes != state_info.size:
        raise ValueError(
            f"Iterator state size, {iterator_state.nbytes} bytes, for iterator type {type(it)} "
            f"does not match expected size, {state_info.size} bytes"
        )
    alignment = state_info.alignment

    # Compile advance operation
    advance_abi_name = f"advance_{_get_abi_suffix()}"
    wrapped_advance, wrapper_sig = create_advance_void_ptr_wrapper(
        it.advance, state_ptr_type
    )
    advance_ltoir, _ = _cached_compile(
        wrapped_advance, wrapper_sig, abi_name=advance_abi_name, output="ltoir"
    )
    advance_op = Op(
        operator_type=OpKind.STATELESS,
        name=advance_abi_name,
        ltoir=advance_ltoir,
    )

    # Compile dereference operation based on io_kind
    if io_kind == "input":
        deref_abi_name = f"input_dereference_{_get_abi_suffix()}"
        wrapped_deref, wrapper_sig = create_input_dereference_void_ptr_wrapper(
            it.input_dereference, state_ptr_type, numba_value_type
        )
    elif io_kind == "output":
        deref_abi_name = f"output_dereference_{_get_abi_suffix()}"
        wrapped_deref, wrapper_sig = create_output_dereference_void_ptr_wrapper(
            it.output_dereference, state_ptr_type, numba_value_type
        )
    else:
        raise ValueError(f"Invalid io_kind: {io_kind}. Must be 'input' or 'output'")

    deref_ltoir, _ = _cached_compile(
        wrapped_deref, wrapper_sig, abi_name=deref_abi_name, output="ltoir"
    )
    deref_op = Op(
        operator_type=OpKind.STATELESS,
        name=deref_abi_name,
        ltoir=deref_ltoir,
    )

    # Get TypeInfo directly from TypeDescriptor
    value_type_info = it.value_type.info()

    return Iterator(
        alignment,
        IteratorKind.ITERATOR,
        advance_op,
        deref_op,
        value_type_info,
        state=it.state,
    )


# -----------------------------------------------------------------------------
# Stateless ops
# -----------------------------------------------------------------------------


def _compile_op(op, input_types, output_type=None):
    """Compile a user-provided stateless operator for use with CCCL algorithms."""
    from . import types as cccl_types
    from ._bindings import Op, OpKind
    from ._odr_helpers import create_op_void_ptr_wrapper

    # Ensure any gpu_struct classes referenced in the op are registered
    _ensure_function_structs_registered(op)

    numba_input_types = tuple(
        type_descriptor_to_numba(t) if isinstance(t, cccl_types.TypeDescriptor) else t
        for t in input_types
    )

    if isinstance(output_type, cccl_types.TypeDescriptor):
        numba_output_type = type_descriptor_to_numba(output_type)
    else:
        numba_output_type = output_type

    sig = numba_output_type(*numba_input_types)
    wrapped_op, wrapper_sig = create_op_void_ptr_wrapper(op, sig)
    ltoir, _ = numba.cuda.compile(wrapped_op, sig=wrapper_sig, output="ltoir")
    return Op(
        operator_type=OpKind.STATELESS,
        name=wrapped_op.__name__,
        ltoir=ltoir,
        state_alignment=1,
        state=None,
    )


class _StatelessOp(OpAdapter):
    """Adapter for stateless callables."""

    __slots__ = ["_func", "_cachable"]

    def __init__(self, func: Callable):
        self._func = func
        self._cachable = CachableFunction(func)

    def compile(self, input_types, output_type=None) -> Op:
        return _compile_op(self._func, input_types, output_type)

    @property
    def func(self) -> Callable:
        """Access the wrapped callable."""
        return self._func


# -----------------------------------------------------------------------------
# Stateful ops
# -----------------------------------------------------------------------------
#
# Stateful ops are callables that capture device arrays (state) as globals or
# closure variables.
#
# When numba-cuda encounters a device function that references device
# arrays in its globals/closures, it captures pointers to the
# referenced arrays as constants.  If an object being referenced
# changes (e.g., common in a loop) the device function will be
# recompiled because it sees a new pointer.
#
# We avoid this by relying on support for stateful ops in the CCCL C
# library itself. Here's how this works
#
# 1. Detect any global/closure device arrays referenced by a given callable.
#
# 2. Transform the callable's AST to add parameters for each detected device array.
#
# 3. Compilation: the compiled device function whose LTOIR we
#    eventually pass to the CCCL C library must take a single `void*`
#    argument for is state.  The corresponding declaration of stateful
#    operators in CCCL C looks like:
#
#       extern "C" __device__ void foo(void* state, )`.
#
#    Thus, we have Thus, at compilation time, we
#    define some intrinsics to unpack a `void*` into typed arrays and
#    invoke the transformed callable from step 2.  Much of this is
#    implemented in _odr_helpers.py


def _detect_device_array_globals(func: Callable) -> List[Tuple[str, object]]:
    """
        Detect device arrays referenced as globals in a function.
    }
        Args:
            func: The function to inspect

        Returns:
            List of (name, array) tuples for detected device arrays
    """
    state_arrays = []
    code = func.__code__

    for name in code.co_names:
        val = func.__globals__.get(name)
        if val is not None and is_device_array(val):
            state_arrays.append((name, val))

    return state_arrays


def _detect_device_array_closures(func: Callable) -> List[Tuple[str, object]]:
    """
    Detect device arrays captured in function closures.

    Args:
        func: The function to inspect

    Returns:
        List of (name, array) tuples for detected device arrays
    """
    state_arrays: List[Tuple[str, object]] = []
    code = func.__code__
    closure = func.__closure__

    if closure is None:
        return state_arrays

    # co_freevars contains the names of closure variables
    for name, cell in zip(code.co_freevars, closure):
        try:
            val = cell.cell_contents
            if is_device_array(val):
                state_arrays.append((name, val))
        except ValueError:
            # Cell is empty
            pass

    return state_arrays


def _detect_all_device_arrays(func: Callable) -> List[Tuple[str, object]]:
    """
    Detect all device arrays referenced by a function (globals + closures).

    Args:
        func: The function to inspect

    Returns:
        List of (name, array) tuples for detected device arrays
    """
    globals_arrays = _detect_device_array_globals(func)
    closure_arrays = _detect_device_array_closures(func)
    return globals_arrays + closure_arrays


class _AddStateParameters(ast.NodeTransformer):
    """AST transformer that adds state parameters to a function definition."""

    def __init__(self, state_names: List[str]):
        self.state_names = state_names

    def visit_FunctionDef(self, node: ast.FunctionDef) -> ast.FunctionDef:
        # Prepend state parameters to the function arguments
        # Inner function signature: (state_arrays..., regular_args...)
        new_args = [ast.arg(arg=name, annotation=None) for name in self.state_names]
        node.args.args = new_args + node.args.args
        return node

    def visit_AsyncFunctionDef(
        self, node: ast.AsyncFunctionDef
    ) -> ast.AsyncFunctionDef:
        # Handle async functions the same way
        new_args = [ast.arg(arg=name, annotation=None) for name in self.state_names]
        node.args.args = new_args + node.args.args
        return node


def _transform_function_ast(func: Callable, state_names: List[str]) -> Callable:
    """
    Transform a function to add state arrays captured as globals or closures
    as explicit parameters.

    For example, if the function is:

        def func(x): return x + state[0]  # state is a global array

    Then the transformed function will be:

        def func(state, x): return x + state[0]

    Args:
        func: The original function
        state_names: Names of device arrays to add as parameters

    Returns:
        A new function with state arrays as explicit parameters,
        appearing after the regular parameters.

    Raises:
        ValueError: If the function source cannot be obtained
    """
    # Get source code
    try:
        source = inspect.getsource(func)
    except (OSError, TypeError) as e:
        raise ValueError(
            f"Cannot get source code for function '{func.__name__}'. "
        ) from e

    # Dedent source (in case function is defined inside a class/function)
    source = textwrap.dedent(source)

    # Parse to AST
    tree = ast.parse(source)

    # Transform: add state parameters
    transformer = _AddStateParameters(state_names)
    tree = transformer.visit(tree)
    ast.fix_missing_locations(tree)

    # Compile and execute to create new function
    # We need to provide the original function's globals for imports
    # AND inject closure variables so they're accessible in the new function
    globals_dict = func.__globals__.copy()

    # Inject closure variables (except state arrays which become parameters)
    if func.__closure__:
        for name, cell in zip(func.__code__.co_freevars, func.__closure__):
            if name not in state_names:  # Don't inject state arrays
                try:
                    globals_dict[name] = cell.cell_contents
                except ValueError:
                    pass  # Cell is empty

    local_ns: dict[str, Callable] = {}
    exec(
        compile(tree, filename=f"<auto_stateful:{func.__name__}>", mode="exec"),
        globals_dict,
        local_ns,
    )

    # Get the transformed function
    transformed_func = local_ns[func.__name__]

    return transformed_func


def _extract_state(func: Callable):
    # Detect device arrays
    state_info = _detect_all_device_arrays(func)

    # Extract names and arrays
    state_names = [name for name, _ in state_info]
    state_arrays: List[DeviceArrayLike] = [arr for _, arr in state_info]  # type: ignore[misc]

    return state_names, state_arrays


def _compile_stateful_op(op, input_types, state_arrays, output_type=None):
    """
    Compile a stateful operator for use with CCCL algorithms.

    Args:
        op: The operator function (already transformed to take state as explicit parameters)
        input_types: Tuple of TypeDescriptors for regular input arguments
        state_arrays: Tuple of device arrays containing state
        output_type: Optional TypeDescriptor for return value (inferred if None)

    Returns:
        Compiled Op object for C++ interop
    """

    # Ensure any gpu_struct classes referenced in the op are registered
    _ensure_function_structs_registered(op)

    # Validate all state arrays are contiguous
    for i, state_array in enumerate(state_arrays):
        if not is_contiguous(state_array):
            raise ValueError(f"state array {i} must be contiguous")

    # Convert input types to Numba types
    numba_input_types = tuple(type_descriptor_to_numba(t) for t in input_types)

    # Create Numba array types for state arrays
    state_array_types = [
        numba.types.Array(numba.from_dtype(get_dtype(s)), 1, "A") for s in state_arrays
    ]

    # Infer output type if needed
    if output_type is None:
        # Compile with Numba to infer return type
        # The transformed function expects (state_arrays..., regular_args...)
        all_numba_input_types = tuple(state_array_types) + numba_input_types
        sanitized_name = sanitize_identifier(op.__name__)
        unique_suffix = hex(id(op))[2:]
        abi_name = f"{sanitized_name}_{unique_suffix}"
        _, return_type = numba.cuda.compile(
            op, all_numba_input_types, abi_info={"abi_name": abi_name}
        )
        # Convert return type to TypeDescriptor
        output_type = cccl_types.from_numpy_dtype(
            numba.np.numpy_support.as_dtype(return_type)
        )

    # Convert output type to Numba type
    numba_output_type = type_descriptor_to_numba(output_type)

    # Build full signature: output_type(state_arrays..., regular_args...)
    sig = numba_output_type(*state_array_types, *numba_input_types)

    # Get state pointers - pointers to the device array data
    state_ptrs = [get_data_pointer(arr) for arr in state_arrays]

    # Get shape and itemsize from each state array
    state_info = []
    for state_array in state_arrays:
        state_info.append(
            {
                "shape": len(state_array),
                "itemsize": get_dtype(state_array).itemsize,
                "strides": get_dtype(state_array).itemsize,
            }
        )

    # All pointers have the same alignment, use pointer-sized int alignment
    state_alignment = np.dtype(np.intp).alignment

    # Create the stateful wrapper (constructs arrays from pointers)
    wrapped_op, wrapper_sig = create_stateful_op_void_ptr_wrapper(
        op, sig, state_array_types, state_info
    )

    # Compile the wrapper to LTOIR
    ltoir, _ = numba.cuda.compile(wrapped_op, sig=wrapper_sig, output="ltoir")

    # Pack all data pointers as bytes (sequentially)
    state_bytes = struct.pack(f"{len(state_ptrs)}P", *state_ptrs)

    # Return Op with STATEFUL kind and packed pointers
    return Op(
        operator_type=OpKind.STATEFUL,
        name=wrapped_op.__name__,
        ltoir=ltoir,
        state_alignment=state_alignment,
        state=state_bytes,
    )


class _JitOpState:
    def __init__(self, names: List[str], arrays: List[DeviceArrayLike]):
        self.names = names
        self.arrays = arrays

    def get_cache_key(self) -> Hashable:
        return (tuple(self.names), tuple(get_dtype(s) for s in self.arrays))

    def to_bytes(self):
        state_ptrs = [get_data_pointer(arr) for arr in self.arrays]
        return struct.pack(f"{len(state_ptrs)}P", *state_ptrs)


class _StatefulOp(OpAdapter):
    __slots__ = ["_func", "_cachable", "_state"]

    def __init__(self, func, state):
        self._func = func
        self._cachable = CachableFunction(self._func)
        self._state = state

    @property
    def state(self):
        return self._state

    def compile(self, input_types, output_type=None) -> Op:
        transformed_func = _transform_function_ast(self._func, self._state.names)
        return _compile_stateful_op(
            transformed_func, input_types, self._state.arrays, output_type
        )

    def get_cache_key(self):
        return (self._func.__name__, self._cachable, self._state.get_cache_key())

    @property
    def is_stateful(self) -> bool:
        return True

    def update_op_state(self, cccl_op) -> None:
        """
        Update state by detecting device arrays from the Python callable.

        Args:
            cccl_op: The compiled CCCL Op to update
            op: The original Python callable (needed to detect current arrays)
        """
        cccl_op.state = self._state.to_bytes()

    @property
    def func(self) -> Callable:
        """Access the wrapped callable."""
        return self._func


def to_jit_op_adapter(op: Callable) -> OpAdapter:
    """
    Convert the Python callable into the appropriate stateful or stateless
    OpAdapter, depending on whether or not it references any global state.
    """
    state_names, state_arrays = _extract_state(op)
    if state_names:
        return _StatefulOp(op, _JitOpState(state_names, state_arrays))
    else:
        return _StatelessOp(op)


cache_with_registered_key_functions.register(_StatelessOp, lambda op: op._cachable)

cache_with_registered_key_functions.register(_StatefulOp, lambda op: op.get_cache_key())


__all__ = [
    "to_jit_op_adapter",
    "compile_iterator",
    "type_descriptor_to_numba",
]



# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from typing import Callable

from ._bindings import Op, OpKind
from ._caching import CachableFunction, cache_with_registered_key_functions


def _is_well_known_op(op: OpKind) -> bool:
    return isinstance(op, OpKind) and op not in (OpKind.STATELESS, OpKind.STATEFUL)


class _OpAdapter:
    """
    Provides a unified interface for operators, whether they are:
    - Well-known operations (OpKind.PLUS, OpKind.MAXIMUM, etc.)
    - Stateless user-provided callables
    - Stateful user-provided callables
    """

    def compile(self, input_types, output_type=None) -> Op:
        """
        Compile this operator to an Op for CCCL interop.

        Args:
            input_types: Tuple of TypeDescriptors for input arguments
            output_type: Optional TypeDescriptor for return value (inferred if None)

        Returns:
            Compiled Op object for C++ interop
        """
        raise NotImplementedError("Subclasses must implement this method")

    @property
    def is_stateful(self) -> bool:
        """Return True if this op has runtime state."""
        return False

    def update_op_state(self, cccl_op: Op) -> None:
        """
        Update the Op's state bytes.

        Args:
            cccl_op: The compiled CCCL Op to update

        Default implementation is a no-op (for stateless ops).
        """
        pass

    @property
    def func(self) -> Callable | None:
        """The underlying callable, if any."""
        return None


class _WellKnownOp(_OpAdapter):
    """Internal wrapper for well-known OpKind values."""

    __slots__ = ["_kind"]

    def __init__(self, kind: OpKind):
        if not _is_well_known_op(kind):
            raise ValueError(
                f"OpKind.{kind.name} is not a well-known operation. "
                "Use OpKind.PLUS, OpKind.MAXIMUM, etc."
            )
        self._kind = kind

    def compile(self, input_types, output_type=None) -> Op:
        return Op(
            operator_type=self._kind,
            name="",
            ltoir=b"",
            state_alignment=1,
            state=b"",
        )

    @property
    def kind(self) -> OpKind:
        """The underlying OpKind."""
        return self._kind


# Public aliases
OpAdapter = _OpAdapter


def make_op_adapter(op) -> OpAdapter:
    """
    Create an Op from a callable or well-known OpKind.

    Args:
        op: Callable or OpKind

    Returns:
        A value with appropriate subtype of _BaseOp
    """
    from ._jit import to_jit_op_adapter

    # Already an _OpAdapter instance:
    if isinstance(op, _OpAdapter):
        return op

    # Well-known operation
    if isinstance(op, OpKind):
        return _WellKnownOp(op)

    # It's a Python callable
    return to_jit_op_adapter(op)


__all__ = [
    "OpAdapter",
    "OpKind",
    "make_op_adapter",
]


cache_with_registered_key_functions.register(
    _WellKnownOp, lambda op: (op._kind.name, op._kind.value)
)

cache_with_registered_key_functions.register(
    OpKind, lambda kind: (kind.name, kind.value)
)

cache_with_registered_key_functions.register(
    type(lambda: None), lambda func: CachableFunction(func)
)
